'use strict';

var luxon = require('luxon');
var perf_hooks = require('perf_hooks');
var uuid = require('uuid');
var backendCommon = require('@backstage/backend-common');
var catalogModel = require('@backstage/catalog-model');
var express = require('express');
var Router = require('express-promise-router');

function _interopDefaultCompat (e) { return e && typeof e === 'object' && 'default' in e ? e : { default: e }; }

var express__default = /*#__PURE__*/_interopDefaultCompat(express);
var Router__default = /*#__PURE__*/_interopDefaultCompat(Router);

const INCREMENTAL_ENTITY_PROVIDER_ANNOTATION = "backstage.io/incremental-provider-name";

class IncrementalIngestionEngine {
  constructor(options) {
    this.options = options;
    this.manager = options.manager;
    this.restLength = luxon.Duration.fromObject(options.restLength);
    this.backoff = options.backoff ?? [
      { minutes: 1 },
      { minutes: 5 },
      { minutes: 30 },
      { hours: 3 }
    ];
  }
  restLength;
  backoff;
  manager;
  async taskFn(signal) {
    try {
      this.options.logger.debug("Begin tick");
      await this.handleNextAction(signal);
    } catch (error) {
      this.options.logger.error(`${error}`);
      throw error;
    } finally {
      this.options.logger.debug("End tick");
    }
  }
  async handleNextAction(signal) {
    await this.options.ready;
    const result = await this.getCurrentAction();
    if (result) {
      const { ingestionId, nextActionAt, nextAction, attempts } = result;
      switch (nextAction) {
        case "rest":
          if (Date.now() > nextActionAt) {
            await this.manager.clearFinishedIngestions(
              this.options.provider.getProviderName()
            );
            this.options.logger.info(
              `incremental-engine: Ingestion ${ingestionId} rest period complete. Ingestion will start again`
            );
            await this.manager.setProviderComplete(ingestionId);
          } else {
            this.options.logger.info(
              `incremental-engine: Ingestion '${ingestionId}' rest period continuing`
            );
          }
          break;
        case "ingest":
          try {
            await this.manager.setProviderBursting(ingestionId);
            const done = await this.ingestOneBurst(ingestionId, signal);
            if (done) {
              this.options.logger.info(
                `incremental-engine: Ingestion '${ingestionId}' complete, transitioning to rest period of ${this.restLength.toHuman()}`
              );
              await this.manager.setProviderResting(
                ingestionId,
                this.restLength
              );
            } else {
              await this.manager.setProviderInterstitial(ingestionId);
              this.options.logger.info(
                `incremental-engine: Ingestion '${ingestionId}' continuing`
              );
            }
          } catch (error) {
            if (error.message && error.message === "CANCEL") {
              this.options.logger.info(
                `incremental-engine: Ingestion '${ingestionId}' canceled`
              );
              await this.manager.setProviderCanceling(
                ingestionId,
                error.message
              );
            } else {
              const currentBackoff = luxon.Duration.fromObject(
                this.backoff[Math.min(this.backoff.length - 1, attempts)]
              );
              const backoffLength = currentBackoff.as("milliseconds");
              this.options.logger.error(error);
              const truncatedError = typeof error === "string" ? error.substring(0, 700) : `${error}`;
              this.options.logger.error(
                `incremental-engine: Ingestion '${ingestionId}' threw an error during ingestion burst. Ingestion will backoff for ${currentBackoff.toHuman()} (${truncatedError})`
              );
              await this.manager.setProviderBackoff(
                ingestionId,
                attempts,
                error,
                backoffLength
              );
            }
          }
          break;
        case "backoff":
          if (Date.now() > nextActionAt) {
            this.options.logger.info(
              `incremental-engine: Ingestion '${ingestionId}' backoff complete, will attempt to resume`
            );
            await this.manager.setProviderIngesting(ingestionId);
          } else {
            this.options.logger.info(
              `incremental-engine: Ingestion '${ingestionId}' backoff continuing`
            );
          }
          break;
        case "cancel":
          this.options.logger.info(
            `incremental-engine: Ingestion '${ingestionId}' canceling, will restart`
          );
          await this.manager.setProviderCanceled(ingestionId);
          break;
        default:
          this.options.logger.error(
            `incremental-engine: Ingestion '${ingestionId}' received unknown action '${nextAction}'`
          );
      }
    } else {
      this.options.logger.error(
        `incremental-engine: Engine tried to create duplicate ingestion record for provider '${this.options.provider.getProviderName()}'.`
      );
    }
  }
  async getCurrentAction() {
    const providerName = this.options.provider.getProviderName();
    const record = await this.manager.getCurrentIngestionRecord(providerName);
    if (record) {
      this.options.logger.info(
        `incremental-engine: Ingestion record found: '${record.id}'`
      );
      return {
        ingestionId: record.id,
        nextAction: record.next_action,
        attempts: record.attempts,
        nextActionAt: record.next_action_at.valueOf()
      };
    }
    const result = await this.manager.createProviderIngestionRecord(
      providerName
    );
    if (result) {
      this.options.logger.info(
        `incremental-engine: Ingestion record created: '${result.ingestionId}'`
      );
    }
    return result;
  }
  async ingestOneBurst(id, signal) {
    const lastMark = await this.manager.getLastMark(id);
    const cursor = lastMark ? lastMark.cursor : void 0;
    let sequence = lastMark ? lastMark.sequence + 1 : 0;
    const start = perf_hooks.performance.now();
    let count = 0;
    let done = false;
    this.options.logger.info(
      `incremental-engine: Ingestion '${id}' burst initiated`
    );
    await this.options.provider.around(async (context) => {
      let next = await this.options.provider.next(context, cursor);
      count++;
      while (true) {
        done = next.done;
        await this.mark(id, sequence, next.entities, next.done, next.cursor);
        if (signal.aborted || next.done) {
          break;
        } else {
          next = await this.options.provider.next(context, next.cursor);
          count++;
          sequence++;
        }
      }
    });
    this.options.logger.info(
      `incremental-engine: Ingestion '${id}' burst complete. (${count} batches in ${Math.round(
        perf_hooks.performance.now() - start
      )}ms).`
    );
    return done;
  }
  async mark(id, sequence, entities, done, cursor) {
    this.options.logger.debug(
      `incremental-engine: Ingestion '${id}': MARK ${entities.length} entities, cursor: ${JSON.stringify(cursor)}, done: ${done}`
    );
    const markId = uuid.v4();
    await this.manager.createMark({
      record: {
        id: markId,
        ingestion_id: id,
        cursor,
        sequence
      }
    });
    if (entities.length > 0) {
      await this.manager.createMarkEntities(markId, entities);
    }
    const added = entities.map((deferred) => ({
      ...deferred,
      entity: {
        ...deferred.entity,
        metadata: {
          ...deferred.entity.metadata,
          annotations: {
            ...deferred.entity.metadata.annotations,
            [INCREMENTAL_ENTITY_PROVIDER_ANNOTATION]: this.options.provider.getProviderName()
          }
        }
      }
    }));
    const removed = done ? [] : await this.manager.computeRemoved(
      this.options.provider.getProviderName(),
      id
    );
    await this.options.connection.applyMutation({
      type: "delta",
      added,
      removed
    });
  }
}

async function applyDatabaseMigrations(knex) {
  const migrationsDir = backendCommon.resolvePackagePath("@frontside/backstage-plugin-incremental-ingestion-backend", "migrations");
  await knex.raw("CREATE SCHEMA IF NOT EXISTS ingestion;");
  await knex.migrate.latest({
    schemaName: "ingestion",
    directory: migrationsDir
  });
}

class IncrementalIngestionDatabaseManager {
  client;
  constructor(options) {
    this.client = options.client;
  }
  /**
   * Performs an update to the ingestion record with matching `id`.
   * @param options - IngestionRecordUpdate
   */
  async updateIngestionRecordById(options) {
    await this.client.transaction(async (tx) => {
      const { ingestionId, update } = options;
      await tx("ingestion.ingestions").where("id", ingestionId).update(update);
    });
  }
  /**
   * Performs an update to the ingestion record with matching provider name. Will only update active records.
   * @param provider - string
   * @param update - Partial<IngestionUpsertIFace>
   */
  async updateIngestionRecordByProvider(provider, update) {
    await this.client.transaction(async (tx) => {
      await tx("ingestion.ingestions").where("provider_name", provider).andWhere("completion_ticket", "open").update(update);
    });
  }
  /**
   * Performs an insert into the `ingestion.ingestions` table with the supplied values.
   * @param record - IngestionUpsertIFace
   */
  async insertIngestionRecord(record) {
    await this.client.transaction(async (tx) => {
      await tx("ingestion.ingestions").insert(record);
    });
  }
  async deleteMarkEntities(tx, ids) {
    const chunks = [];
    for (let i = 0; i < ids.length; i += 100) {
      const chunk = ids.slice(i, i + 100);
      chunks.push(chunk);
    }
    let deleted = 0;
    for (const chunk of chunks) {
      const chunkDeleted = await tx("ingestion.ingestion_mark_entities").delete().whereIn(
        "id",
        chunk.map((entry) => entry.id)
      );
      deleted += chunkDeleted;
    }
    return deleted;
  }
  /**
   * Finds the current ingestion record for the named provider.
   * @param provider - string
   * @returns IngestionRecord | undefined
   */
  async getCurrentIngestionRecord(provider) {
    return await this.client.transaction(async (tx) => {
      const record = await tx("ingestion.ingestions").where("provider_name", provider).andWhere("completion_ticket", "open").first();
      return record;
    });
  }
  /**
   * Removes all entries from `ingestion_marks_entities`, `ingestion_marks`, and `ingestions`
   * for prior ingestions that completed (i.e., have a `completion_ticket` value other than 'open').
   * @param provider - string
   * @returns A count of deletions for each record type.
   */
  async clearFinishedIngestions(provider) {
    return await this.client.transaction(async (tx) => {
      const markEntitiesDeleted = await tx("ingestion.ingestion_mark_entities").delete().whereIn(
        "ingestion_mark_id",
        tx("ingestion.ingestion_marks").select("id").whereIn(
          "ingestion_id",
          tx("ingestion.ingestions").select("id").where("provider_name", provider).andWhereNot("completion_ticket", "open")
        )
      );
      const marksDeleted = await tx("ingestion.ingestion_marks").delete().whereIn(
        "ingestion_id",
        tx("ingestion.ingestions").select("id").where("provider_name", provider).andWhereNot("completion_ticket", "open")
      );
      const ingestionsDeleted = await tx("ingestion.ingestions").delete().where("provider_name", provider).andWhereNot("completion_ticket", "open");
      return {
        deletions: {
          markEntitiesDeleted,
          marksDeleted,
          ingestionsDeleted
        }
      };
    });
  }
  /**
   * Automatically cleans up duplicate ingestion records if they were accidentally created.
   * Any ingestion record where the `rest_completed_at` is null (meaning it is active) AND
   * the ingestionId is incorrect is a duplicate ingestion record.
   * @param ingestionId - string
   * @param provider - string
   */
  async clearDuplicateIngestions(ingestionId, provider) {
    await this.client.transaction(async (tx) => {
      const invalid = await tx("ingestion.ingestions").where("provider_name", provider).andWhere("rest_completed_at", null).andWhereNot("id", ingestionId);
      if (invalid.length > 0) {
        await tx("ingestion.ingestions").delete().whereIn("id", invalid);
        await tx("ingestion.ingestion_mark_entities").delete().whereIn(
          "ingestion_mark_id",
          tx("ingestion.ingestion_marks").select("id").whereIn("ingestion_id", invalid)
        );
        await tx("ingestion.ingestion_marks").delete().whereIn("ingestion_id", invalid);
      }
    });
  }
  /**
   * This method fully purges and resets all ingestion records for the named provider, and
   * leaves it in a paused state.
   * @param provider - string
   * @returns Counts of all deleted ingestion records
   */
  async purgeAndResetProvider(provider) {
    return await this.client.transaction(async (tx) => {
      const ingestionIDs = await tx("ingestion.ingestions").select("id").where("provider_name", provider);
      const markIDs = ingestionIDs.length > 0 ? await tx("ingestion.ingestion_marks").select("id").whereIn(
        "ingestion_id",
        ingestionIDs.map((entry) => entry.id)
      ) : [];
      const markEntityIDs = markIDs.length > 0 ? await tx("ingestion.ingestion_mark_entities").select("id").whereIn(
        "ingestion_mark_id",
        markIDs.map((entry) => entry.id)
      ) : [];
      const markEntitiesDeleted = await this.deleteMarkEntities(
        tx,
        markEntityIDs
      );
      const marksDeleted = markIDs.length > 0 ? await tx("ingestion.ingestion_marks").delete().whereIn(
        "ingestion_id",
        ingestionIDs.map((entry) => entry.id)
      ) : 0;
      const ingestionsDeleted = await tx("ingestion.ingestions").delete().where("provider_name", provider);
      const next_action_at = /* @__PURE__ */ new Date();
      next_action_at.setTime(next_action_at.getTime() + 24 * 60 * 60 * 1e3);
      await this.insertIngestionRecord({
        id: uuid.v4(),
        next_action: "rest",
        provider_name: provider,
        next_action_at,
        ingestion_completed_at: /* @__PURE__ */ new Date(),
        status: "resting",
        completion_ticket: uuid.v4()
      });
      return { provider, ingestionsDeleted, marksDeleted, markEntitiesDeleted };
    });
  }
  /**
   * Creates a new ingestion record.
   * @param provider - string
   * @returns A new ingestion record
   */
  async createProviderIngestionRecord(provider) {
    const ingestionId = uuid.v4();
    const nextAction = "ingest";
    try {
      await this.insertIngestionRecord({
        id: ingestionId,
        next_action: nextAction,
        provider_name: provider,
        status: "bursting",
        completion_ticket: "open"
      });
      return { ingestionId, nextAction, attempts: 0, nextActionAt: Date.now() };
    } catch (_e) {
      return void 0;
    }
  }
  /**
   * Computes which entities to remove, if any, at the end of a burst.
   * @param provider - string
   * @param ingestionId - string
   * @returns All entities to remove for this burst.
   */
  async computeRemoved(provider, ingestionId) {
    return await this.client.transaction(async (tx) => {
      const removed = await tx(
        "final_entities"
      ).select(
        tx.ref("final_entity").as("entity"),
        tx.ref("refresh_state.entity_ref").as("ref")
      ).join(
        "refresh_state",
        "refresh_state.entity_id",
        "final_entities.entity_id"
      ).whereRaw(
        `((final_entity::json #>> '{metadata, annotations, ${INCREMENTAL_ENTITY_PROVIDER_ANNOTATION}}')) = ?`,
        [provider]
      ).whereNotIn(
        "entity_ref",
        tx("ingestion.ingestion_marks").join(
          "ingestion.ingestion_mark_entities",
          "ingestion.ingestion_marks.id",
          "ingestion.ingestion_mark_entities.ingestion_mark_id"
        ).select("ingestion.ingestion_mark_entities.ref").where("ingestion.ingestion_marks.ingestion_id", ingestionId)
      );
      return removed.map((entity) => {
        return { entity: JSON.parse(entity.entity) };
      });
    });
  }
  /**
   * Performs a lookup of all providers that have duplicate active ingestion records.
   * @returns An array of all duplicate active ingestions
   */
  async healthcheck() {
    return await this.client.transaction(async (tx) => {
      const records = await tx(
        "ingestion.ingestions"
      ).distinct("id", "provider_name").where("rest_completed_at", null);
      return records;
    });
  }
  /**
   * Skips any wait time for the next action to run.
   * @param provider - string
   */
  async triggerNextProviderAction(provider) {
    await this.updateIngestionRecordByProvider(provider, {
      next_action_at: /* @__PURE__ */ new Date()
    });
  }
  /**
   * Purges the following tables:
   * * `ingestion.ingestions`
   * * `ingestion.ingestion_marks`
   * * `ingestion.ingestion_mark_entities`
   *
   * This function leaves the ingestions table with all providers in a paused state.
   * @returns Results from cleaning up all ingestion tables.
   */
  async cleanupProviders() {
    const providers = await this.listProviders();
    const ingestionsDeleted = await this.purgeTable("ingestion.ingestions");
    const next_action_at = /* @__PURE__ */ new Date();
    next_action_at.setTime(next_action_at.getTime() + 24 * 60 * 60 * 1e3);
    for (const provider of providers) {
      await this.insertIngestionRecord({
        id: uuid.v4(),
        next_action: "rest",
        provider_name: provider,
        next_action_at,
        ingestion_completed_at: /* @__PURE__ */ new Date(),
        status: "resting",
        completion_ticket: uuid.v4()
      });
    }
    const ingestionMarksDeleted = await this.purgeTable(
      "ingestion.ingestion_marks"
    );
    const markEntitiesDeleted = await this.purgeTable(
      "ingestion.ingestion_mark_entities"
    );
    return { ingestionsDeleted, ingestionMarksDeleted, markEntitiesDeleted };
  }
  /**
   * Configures the current ingestion record to ingest a burst.
   * @param ingestionId - string
   */
  async setProviderIngesting(ingestionId) {
    await this.updateIngestionRecordById({
      ingestionId,
      update: { next_action: "ingest" }
    });
  }
  /**
   * Indicates the provider is currently ingesting a burst.
   * @param ingestionId - string
   */
  async setProviderBursting(ingestionId) {
    await this.updateIngestionRecordById({
      ingestionId,
      update: { status: "bursting" }
    });
  }
  /**
   * Finalizes the current ingestion record to indicate that the post-ingestion rest period is complete.
   * @param ingestionId - string
   */
  async setProviderComplete(ingestionId) {
    await this.updateIngestionRecordById({
      ingestionId,
      update: {
        next_action: "nothing (done)",
        rest_completed_at: /* @__PURE__ */ new Date(),
        status: "complete",
        completion_ticket: uuid.v4()
      }
    });
  }
  /**
   * Marks ingestion as complete and starts the post-ingestion rest cycle.
   * @param ingestionId - string
   * @param restLength - Duration
   */
  async setProviderResting(ingestionId, restLength) {
    await this.updateIngestionRecordById({
      ingestionId,
      update: {
        next_action: "rest",
        next_action_at: new Date(Date.now() + restLength.as("milliseconds")),
        ingestion_completed_at: /* @__PURE__ */ new Date(),
        status: "resting"
      }
    });
  }
  /**
   * Marks ingestion as paused after a burst completes.
   * @param ingestionId - string
   */
  async setProviderInterstitial(ingestionId) {
    await this.updateIngestionRecordById({
      ingestionId,
      update: { attempts: 0, status: "interstitial" }
    });
  }
  /**
   * Starts the cancel process for the current ingestion.
   * @param ingestionId - string
   * @param message - string (optional)
   */
  async setProviderCanceling(ingestionId, message) {
    const update = {
      next_action: "cancel",
      last_error: message ? message : void 0,
      next_action_at: /* @__PURE__ */ new Date(),
      status: "canceling"
    };
    await this.updateIngestionRecordById({ ingestionId, update });
  }
  /**
   * Completes the cancel process and triggers a new ingestion.
   * @param ingestionId - string
   */
  async setProviderCanceled(ingestionId) {
    await this.updateIngestionRecordById({
      ingestionId,
      update: {
        next_action: "nothing (canceled)",
        rest_completed_at: /* @__PURE__ */ new Date(),
        status: "complete",
        completion_ticket: uuid.v4()
      }
    });
  }
  /**
   * Configures the current ingestion to wait and retry, due to a data source error.
   * @param ingestionId - string
   * @param attempts - number
   * @param error - Error
   * @param backoffLength - number
   */
  async setProviderBackoff(ingestionId, attempts, error, backoffLength) {
    await this.updateIngestionRecordById({
      ingestionId,
      update: {
        next_action: "backoff",
        attempts: attempts + 1,
        last_error: String(error),
        next_action_at: new Date(Date.now() + backoffLength),
        status: "backing off"
      }
    });
  }
  /**
   * Returns the last record from `ingestion.ingestion_marks` for the supplied ingestionId.
   * @param ingestionId - string
   * @returns MarkRecord | undefined
   */
  async getLastMark(ingestionId) {
    return await this.client.transaction(async (tx) => {
      const mark = await tx("ingestion.ingestion_marks").where("ingestion_id", ingestionId).orderBy("sequence", "desc").first();
      return mark;
    });
  }
  async getAllMarks(ingestionId) {
    return await this.client.transaction(async (tx) => {
      const marks = await tx("ingestion.ingestion_marks").where("ingestion_id", ingestionId).orderBy("sequence", "desc");
      return marks;
    });
  }
  /**
   * Performs an insert into the `ingestion.ingestion_marks` table with the supplied values.
   * @param options - MarkRecordInsert
   */
  async createMark(options) {
    const { record } = options;
    await this.client.transaction(async (tx) => {
      await tx("ingestion.ingestion_marks").insert(record);
    });
  }
  /**
   * Performs an upsert to the `ingestion.ingestion_mark_entities` table for all deferred entities.
   * @param markId - string
   * @param entities - DeferredEntity[]
   */
  async createMarkEntities(markId, entities) {
    await this.client.transaction(async (tx) => {
      await tx("ingestion.ingestion_mark_entities").insert(
        entities.map((entity) => ({
          id: uuid.v4(),
          ingestion_mark_id: markId,
          ref: catalogModel.stringifyEntityRef(entity.entity)
        }))
      );
    });
  }
  /**
   * Deletes the entire content of a table, and returns the number of records deleted.
   * @param table - string
   * @returns number
   */
  async purgeTable(table) {
    return await this.client.transaction(async (tx) => {
      return await tx(table).delete();
    });
  }
  /**
   * Returns a list of all providers.
   * @returns string[]
   */
  async listProviders() {
    return await this.client.transaction(async (tx) => {
      const providers = await tx(
        "ingestion.ingestions"
      ).distinct("provider_name");
      return providers.map((entry) => entry.provider_name);
    });
  }
  async updateByName(provider, update) {
    await this.updateIngestionRecordByProvider(provider, update);
  }
}

const createIncrementalProviderRouter = async (manager, logger) => {
  const router = Router__default.default();
  router.use(express__default.default.json());
  router.get("/health", async (_, res) => {
    const records = await manager.healthcheck();
    const providers = records.map((record) => record.provider_name);
    const duplicates = [
      ...new Set(providers.filter((e, i, a) => a.indexOf(e) !== i))
    ];
    if (duplicates.length > 0) {
      res.json({ healthy: false, duplicateIngestions: duplicates });
    } else {
      res.json({ healthy: true });
    }
  });
  router.delete("/cleanup", async (_, res) => {
    const result = await manager.cleanupProviders();
    res.json(result);
  });
  router.get("/:provider", async (req, res) => {
    const { provider } = req.params;
    const record = await manager.getCurrentIngestionRecord(provider);
    if (record) {
      res.json({
        success: true,
        status: {
          current_action: record.status,
          next_action_at: new Date(record.next_action_at)
        },
        last_error: record.last_error
      });
    } else {
      const providers = await manager.listProviders();
      if (providers.includes(provider)) {
        res.json({
          success: true,
          status: {
            current_action: "rest complete, waiting to start"
          }
        });
      } else {
        logger.error(
          `${provider} - No ingestion record found in the database!`
        );
        res.status(404).json({
          success: false,
          status: {},
          last_error: `Provider '${provider}' not found`
        });
      }
    }
  });
  router.put("/:provider", async (req, res) => {
    const { provider } = req.params;
    const record = await manager.getCurrentIngestionRecord(provider);
    if (record) {
      await manager.triggerNextProviderAction(provider);
      res.json({
        success: true,
        message: `${provider}: Next action triggered.`
      });
    } else {
      const providers = await manager.listProviders();
      if (providers.includes(provider)) {
        logger.debug(`${provider} - Ingestion record found`);
        res.json({
          success: true,
          message: "Unable to trigger next action (provider is restarting)"
        });
      } else {
        res.status(404).json({
          success: false,
          message: `Provider '${provider}' not found`
        });
      }
    }
  });
  router.post("/:provider", async (req, res) => {
    const { provider } = req.params;
    const record = await manager.getCurrentIngestionRecord(provider);
    if (record) {
      await manager.updateByName(provider, {
        next_action: "nothing (done)",
        ingestion_completed_at: /* @__PURE__ */ new Date(),
        rest_completed_at: /* @__PURE__ */ new Date(),
        status: "complete"
      });
      res.json({
        success: true,
        message: `${provider}: Next cycle triggered.`
      });
    } else {
      const providers = await manager.listProviders();
      if (providers.includes(provider)) {
        logger.debug(`${provider} - Ingestion record found`);
        res.json({
          success: true,
          message: "Provider is already restarting"
        });
      } else {
        res.status(404).json({
          success: false,
          message: `Provider '${provider}' not found`
        });
      }
    }
  });
  router.post("/:provider/cancel", async (req, res) => {
    const { provider } = req.params;
    const record = await manager.getCurrentIngestionRecord(provider);
    if (record) {
      const next_action_at = /* @__PURE__ */ new Date();
      next_action_at.setTime(next_action_at.getTime() + 24 * 60 * 60 * 1e3);
      await manager.updateByName(provider, {
        next_action: "nothing (done)",
        ingestion_completed_at: /* @__PURE__ */ new Date(),
        next_action_at,
        status: "resting"
      });
      res.json({
        success: true,
        message: `${provider}: Current ingestion canceled.`
      });
    } else {
      const providers = await manager.listProviders();
      if (providers.includes(provider)) {
        logger.debug(`${provider} - Ingestion record found`);
        res.json({
          success: true,
          message: "Provider is currently restarting, please wait."
        });
      } else {
        res.status(404).json({
          success: false,
          message: `Provider '${provider}' not found`
        });
      }
    }
  });
  router.delete("/:provider", async (req, res) => {
    const { provider } = req.params;
    const result = await manager.purgeAndResetProvider(provider);
    res.json(result);
  });
  router.get("/:provider/marks", async (req, res) => {
    const { provider } = req.params;
    const record = await manager.getCurrentIngestionRecord(provider);
    if (record) {
      const id = record.id;
      const records = await manager.getAllMarks(id);
      res.json({ success: true, records });
    } else {
      const providers = await manager.listProviders();
      if (providers.includes(provider)) {
        logger.debug(`${provider} - Ingestion record found`);
        res.json({
          success: true,
          message: "No records yet (provider is restarting)"
        });
      } else {
        logger.error(
          `${provider} - No ingestion record found in the database!`
        );
        res.status(404).json({
          success: false,
          status: {},
          last_error: `Provider '${provider}' not found`
        });
      }
    }
  });
  router.delete("/:provider/marks", async (req, res) => {
    const { provider } = req.params;
    const deletions = await manager.clearFinishedIngestions(provider);
    res.json({
      success: true,
      message: `Expired marks for provider '${provider}' removed.`,
      deletions
    });
  });
  router.use(backendCommon.errorHandler());
  return router;
};

class Deferred {
  resolve;
  reject;
  then;
  catch;
  finally;
  constructor() {
    const promise = new Promise((resolve, reject) => {
      this.resolve = resolve;
      this.reject = reject;
    });
    this.then = promise.then.bind(promise);
    this.catch = promise.catch.bind(promise);
    this.finally = promise.finally.bind(promise);
  }
  [Symbol.toStringTag] = "Deferred";
}
class IncrementalCatalogBuilder {
  constructor(env, builder, client, manager) {
    this.env = env;
    this.builder = builder;
    this.client = client;
    this.manager = manager;
    this.ready = new Deferred();
  }
  /**
   * Creates the incremental catalog builder, which extends the regular catalog builder.
   * @param env - PluginEnvironment
   * @param builder - CatalogBuilder
   * @returns IncrementalCatalogBuilder
   */
  static async create(env, builder) {
    const client = await env.database.getClient();
    const manager = new IncrementalIngestionDatabaseManager({ client });
    return new IncrementalCatalogBuilder(env, builder, client, manager);
  }
  ready;
  async build() {
    await applyDatabaseMigrations(this.client);
    this.ready.resolve();
    const routerLogger = this.env.logger.child({
      router: "IncrementalProviderAdmin"
    });
    const incrementalAdminRouter = await createIncrementalProviderRouter(
      this.manager,
      routerLogger
    );
    return { incrementalAdminRouter, manager: this.manager };
  }
  addIncrementalEntityProvider(provider, options) {
    const { burstInterval, burstLength, restLength } = options;
    const { logger: catalogLogger, database, scheduler } = this.env;
    const ready = this.ready;
    this.builder.addEntityProvider({
      getProviderName: provider.getProviderName.bind(provider),
      async connect(connection) {
        const logger = catalogLogger.child({
          entityProvider: provider.getProviderName()
        });
        logger.info(`Connecting`);
        const client = await database.getClient();
        const manager = new IncrementalIngestionDatabaseManager({ client });
        const engine = new IncrementalIngestionEngine({
          ...options,
          ready,
          manager,
          logger,
          provider,
          restLength,
          connection
        });
        const frequency = luxon.Duration.isDuration(burstInterval) ? burstInterval : luxon.Duration.fromObject(burstInterval);
        const length = luxon.Duration.isDuration(burstLength) ? burstLength : luxon.Duration.fromObject(burstLength);
        await scheduler.scheduleTask({
          id: provider.getProviderName(),
          fn: engine.taskFn.bind(engine),
          frequency,
          timeout: length
        });
      }
    });
  }
}

exports.IncrementalCatalogBuilder = IncrementalCatalogBuilder;
exports.IncrementalIngestionDatabaseManager = IncrementalIngestionDatabaseManager;
//# sourceMappingURL=index.cjs.js.map
