import * as express from 'express';
import { PluginDatabaseManager, UrlReader } from '@backstage/backend-common';
import { PluginTaskScheduler, TaskFunction } from '@backstage/backend-tasks';
import { Config } from '@backstage/config';
import { DeferredEntity, EntityProviderConnection, CatalogBuilder } from '@backstage/plugin-catalog-backend';
import { PermissionAuthorizer } from '@backstage/plugin-permission-common';
import { Duration, DurationObjectUnits } from 'luxon';
import { Logger } from 'winston';
import { Knex } from 'knex';

/** @public */
declare class IncrementalIngestionDatabaseManager {
    private client;
    constructor(options: {
        client: Knex;
    });
    /**
     * Performs an update to the ingestion record with matching `id`.
     * @param options - IngestionRecordUpdate
     */
    updateIngestionRecordById(options: IngestionRecordUpdate): Promise<void>;
    /**
     * Performs an update to the ingestion record with matching provider name. Will only update active records.
     * @param provider - string
     * @param update - Partial<IngestionUpsertIFace>
     */
    updateIngestionRecordByProvider(provider: string, update: Partial<IngestionUpsertIFace>): Promise<void>;
    /**
     * Performs an insert into the `ingestion.ingestions` table with the supplied values.
     * @param record - IngestionUpsertIFace
     */
    insertIngestionRecord(record: IngestionUpsertIFace): Promise<void>;
    private deleteMarkEntities;
    /**
     * Finds the current ingestion record for the named provider.
     * @param provider - string
     * @returns IngestionRecord | undefined
     */
    getCurrentIngestionRecord(provider: string): Promise<IngestionRecord | undefined>;
    /**
     * Removes all entries from `ingestion_marks_entities`, `ingestion_marks`, and `ingestions`
     * for prior ingestions that completed (i.e., have a `completion_ticket` value other than 'open').
     * @param provider - string
     * @returns A count of deletions for each record type.
     */
    clearFinishedIngestions(provider: string): Promise<{
        deletions: {
            markEntitiesDeleted: number;
            marksDeleted: number;
            ingestionsDeleted: number;
        };
    }>;
    /**
     * Automatically cleans up duplicate ingestion records if they were accidentally created.
     * Any ingestion record where the `rest_completed_at` is null (meaning it is active) AND
     * the ingestionId is incorrect is a duplicate ingestion record.
     * @param ingestionId - string
     * @param provider - string
     */
    clearDuplicateIngestions(ingestionId: string, provider: string): Promise<void>;
    /**
     * This method fully purges and resets all ingestion records for the named provider, and
     * leaves it in a paused state.
     * @param provider - string
     * @returns Counts of all deleted ingestion records
     */
    purgeAndResetProvider(provider: string): Promise<{
        provider: string;
        ingestionsDeleted: number;
        marksDeleted: number;
        markEntitiesDeleted: number;
    }>;
    /**
     * Creates a new ingestion record.
     * @param provider - string
     * @returns A new ingestion record
     */
    createProviderIngestionRecord(provider: string): Promise<{
        ingestionId: string;
        nextAction: string;
        attempts: number;
        nextActionAt: number;
    } | undefined>;
    /**
     * Computes which entities to remove, if any, at the end of a burst.
     * @param provider - string
     * @param ingestionId - string
     * @returns All entities to remove for this burst.
     */
    computeRemoved(provider: string, ingestionId: string): Promise<{
        entity: any;
    }[]>;
    /**
     * Performs a lookup of all providers that have duplicate active ingestion records.
     * @returns An array of all duplicate active ingestions
     */
    healthcheck(): Promise<Pick<{
        id: string;
        provider_name: string;
    }, "id" | "provider_name">[]>;
    /**
     * Skips any wait time for the next action to run.
     * @param provider - string
     */
    triggerNextProviderAction(provider: string): Promise<void>;
    /**
     * Purges the following tables:
     * * `ingestion.ingestions`
     * * `ingestion.ingestion_marks`
     * * `ingestion.ingestion_mark_entities`
     *
     * This function leaves the ingestions table with all providers in a paused state.
     * @returns Results from cleaning up all ingestion tables.
     */
    cleanupProviders(): Promise<{
        ingestionsDeleted: void;
        ingestionMarksDeleted: void;
        markEntitiesDeleted: void;
    }>;
    /**
     * Configures the current ingestion record to ingest a burst.
     * @param ingestionId - string
     */
    setProviderIngesting(ingestionId: string): Promise<void>;
    /**
     * Indicates the provider is currently ingesting a burst.
     * @param ingestionId - string
     */
    setProviderBursting(ingestionId: string): Promise<void>;
    /**
     * Finalizes the current ingestion record to indicate that the post-ingestion rest period is complete.
     * @param ingestionId - string
     */
    setProviderComplete(ingestionId: string): Promise<void>;
    /**
     * Marks ingestion as complete and starts the post-ingestion rest cycle.
     * @param ingestionId - string
     * @param restLength - Duration
     */
    setProviderResting(ingestionId: string, restLength: Duration): Promise<void>;
    /**
     * Marks ingestion as paused after a burst completes.
     * @param ingestionId - string
     */
    setProviderInterstitial(ingestionId: string): Promise<void>;
    /**
     * Starts the cancel process for the current ingestion.
     * @param ingestionId - string
     * @param message - string (optional)
     */
    setProviderCanceling(ingestionId: string, message?: string): Promise<void>;
    /**
     * Completes the cancel process and triggers a new ingestion.
     * @param ingestionId - string
     */
    setProviderCanceled(ingestionId: string): Promise<void>;
    /**
     * Configures the current ingestion to wait and retry, due to a data source error.
     * @param ingestionId - string
     * @param attempts - number
     * @param error - Error
     * @param backoffLength - number
     */
    setProviderBackoff(ingestionId: string, attempts: number, error: Error, backoffLength: number): Promise<void>;
    /**
     * Returns the last record from `ingestion.ingestion_marks` for the supplied ingestionId.
     * @param ingestionId - string
     * @returns MarkRecord | undefined
     */
    getLastMark(ingestionId: string): Promise<MarkRecord | undefined>;
    getAllMarks(ingestionId: string): Promise<MarkRecord[]>;
    /**
     * Performs an insert into the `ingestion.ingestion_marks` table with the supplied values.
     * @param options - MarkRecordInsert
     */
    createMark(options: MarkRecordInsert): Promise<void>;
    /**
     * Performs an upsert to the `ingestion.ingestion_mark_entities` table for all deferred entities.
     * @param markId - string
     * @param entities - DeferredEntity[]
     */
    createMarkEntities(markId: string, entities: DeferredEntity[]): Promise<void>;
    /**
     * Deletes the entire content of a table, and returns the number of records deleted.
     * @param table - string
     * @returns number
     */
    purgeTable(table: string): Promise<void>;
    /**
     * Returns a list of all providers.
     * @returns string[]
     */
    listProviders(): Promise<string[]>;
    updateByName(provider: string, update: Partial<IngestionUpsertIFace>): Promise<void>;
}

/**
 * Ingest entities into the catalog in bite-sized chunks.
 *
 * A Normal `EntityProvider` allows you to introduce entities into the
 * processing pipeline by calling an `applyMutation()` on the full set
 * of entities. However, this is not great when the number of entities
 * that you have to keep track of is extremely large because it
 * entails having all of them in memory at once. An
 * `IncrementalEntityProvider` by contrast allows you to provide
 * batches of entities in sequence so that you never need to have more
 * than a few hundred in memory at a time.
 *
 * @public
 */
interface IncrementalEntityProvider<TCursor, TContext> {
    /**
     * This name must be unique between all of the entity providers
     * operating in the catalog.
     */
    getProviderName(): string;
    /**
     * Return a single page of entities from a specific point in the
     * ingestion.
     *
     * @param context - anything needed in order to fetch a single page.
     * @param cursor - a uniqiue value identifying the page to ingest.
     * @returns The entities to be ingested, as well as the cursor of
     * the next page after this one.
     */
    next(context: TContext, cursor?: TCursor): Promise<EntityIteratorResult<TCursor>>;
    /**
     * Do any setup and teardown necessary in order to provide the
     * context for fetching pages. This should always invoke `burst` in
     * order to fetch the individual pages.
     *
     * @param burst - a function which performs a series of iterations
     */
    around(burst: (context: TContext) => Promise<void>): Promise<void>;
}
/**
 * Value returned by an {@link IncrementalEntityProvider} to provide a
 * single page of entities to ingest.
 *
 * @public
 */
interface EntityIteratorResult<T> {
    /**
     * Indicates whether there are any further pages of entities to
     * ingest after this one.
     */
    done: boolean;
    /**
     * A value that marks the page of entities after this one. It will
     * be used to pass into the following invocation of `next()`
     */
    cursor: T;
    /**
     * The entities to ingest.
     */
    entities: DeferredEntity[];
}
/** @public */
interface IncrementalEntityProviderOptions {
    /**
     * Entities are ingested in bursts. This interval determines how
     * much time to wait in between each burst.
     */
    burstInterval: DurationObjectUnits;
    /**
     * Entities are ingested in bursts. This value determines how long
     * to keep ingesting within each burst.
     */
    burstLength: DurationObjectUnits;
    /**
     * After a successful ingestion, the incremental entity provider
     * will rest for this period of time before starting to ingest
     * again.
     */
    restLength: DurationObjectUnits;
    /**
     * In the event of an error during an ingestion burst, the backoff
     * determines how soon it will be retried. E.g.
     * `[{ minutes: 1}, { minutes: 5}, {minutes: 30 }, { hours: 3 }]`
     */
    backoff?: DurationObjectUnits[];
}
/** @public */
type PluginEnvironment = {
    logger: Logger;
    database: PluginDatabaseManager;
    scheduler: PluginTaskScheduler;
    config: Config;
    reader: UrlReader;
    permissions: PermissionAuthorizer;
};
/** @public */
interface IterationEngine {
    taskFn: TaskFunction;
}
/** @public */
interface IterationEngineOptions {
    logger: Logger;
    connection: EntityProviderConnection;
    manager: IncrementalIngestionDatabaseManager;
    provider: IncrementalEntityProvider<unknown, unknown>;
    restLength: DurationObjectUnits;
    ready: Promise<void>;
    backoff?: IncrementalEntityProviderOptions['backoff'];
}
/**
 * The shape of data inserted into or updated in the `ingestion.ingestions` table.
 *
 * @public
 */
interface IngestionUpsertIFace {
    /**
     * The ingestion record id.
     */
    id?: string;
    /**
     * The next action the incremental entity provider will take.
     */
    next_action: 'rest' | 'ingest' | 'backoff' | 'cancel' | 'nothing (done)' | 'nothing (canceled)';
    /**
     * Current status of the incremental entity provider.
     */
    status: 'complete' | 'bursting' | 'resting' | 'canceling' | 'interstitial' | 'backing off';
    /**
     * The name of the incremental entity provider being updated.
     */
    provider_name: string;
    /**
     * Date/time stamp for when the next action will trigger.
     */
    next_action_at?: Date;
    /**
     * A record of the last error generated by the incremental entity provider.
     */
    last_error?: string | null;
    /**
     * The number of attempts the provider has attempted during the current cycle.
     */
    attempts?: number;
    /**
     * Date/time stamp for the completion of ingestion.
     */
    ingestion_completed_at?: Date | string | null;
    /**
     * Date/time stamp for the end of the rest cycle before the next ingestion.
     */
    rest_completed_at?: Date | string | null;
    /**
     * A record of the finalized status of the ingestion record. Values are either 'open' or a uuid.
     */
    completion_ticket: string;
}
/**
 * This interface is for updating an existing ingestion record.
 *
 * @public
 */
interface IngestionRecordUpdate {
    ingestionId: string;
    update: Partial<IngestionUpsertIFace>;
}
/**
 * The expected response from the `ingestion.ingestion_marks` table.
 *
 * @public
 */
interface MarkRecord {
    id: string;
    sequence: number;
    ingestion_id: string;
    cursor: string;
    created_at: string;
}
/**
 * The expected response from the `ingestion.ingestions` table.
 *
 * @public
 */
interface IngestionRecord extends IngestionUpsertIFace {
    id: string;
    next_action_at: Date;
    /**
     * The date/time the ingestion record was created.
     */
    created_at: string;
}
/**
 * This interface supplies all the values for adding an ingestion mark.
 *
 * @public
 */
interface MarkRecordInsert {
    record: {
        id: string;
        ingestion_id: string;
        cursor: unknown;
        sequence: number;
    };
}

/** @public */
declare class IncrementalCatalogBuilder {
    private env;
    private builder;
    private client;
    private manager;
    /**
     * Creates the incremental catalog builder, which extends the regular catalog builder.
     * @param env - PluginEnvironment
     * @param builder - CatalogBuilder
     * @returns IncrementalCatalogBuilder
     */
    static create(env: PluginEnvironment, builder: CatalogBuilder): Promise<IncrementalCatalogBuilder>;
    private ready;
    private constructor();
    build(): Promise<{
        incrementalAdminRouter: express.Router;
        manager: IncrementalIngestionDatabaseManager;
    }>;
    addIncrementalEntityProvider<T, C>(provider: IncrementalEntityProvider<T, C>, options: IncrementalEntityProviderOptions): void;
}

export { type EntityIteratorResult, IncrementalCatalogBuilder, type IncrementalEntityProvider, type IncrementalEntityProviderOptions, IncrementalIngestionDatabaseManager, type IngestionRecord, type IngestionRecordUpdate, type IngestionUpsertIFace, type IterationEngine, type IterationEngineOptions, type MarkRecordInsert, type PluginEnvironment };
